classdef truss7BarMultipleFailures
    properties(GetAccess = 'public', SetAccess = 'private')
        myMaxDeformationNominal;
        myMaxStressNominal;
        myDeformationSafetyFactor;
        myStressSafetyFactor;
        myAreaNominal;
        myAreaStdDev;
    end
    methods
        function obj = truss7BarMultipleFailures 
            % Define constants used within this class
            obj.myDeformationSafetyFactor = 1.06;
            obj.myStressSafetyFactor = 1.08;
            obj.myAreaNominal = 1e-6;
            obj.myAreaStdDev = 2.5e-8;
            
            t = obj.truss7Bar(obj.myAreaNominal); % default area
            t = t.assemble();
            t = t.solve();
            obj.myMaxDeformationNominal = t.myDeformationMax;
            obj.myMaxStressNominal = t.myStressMax;
            
        end
        function [Pf_SE]  = pFailureDeformationOrStressSE(obj,nTotalSamples)
            [betaDeformation] = HansoferLindND(@(zVec)obj.gDeformationWithGradient(zVec),7);
            [betaStress] = HansoferLindND(@(zVec)obj.gStressWithGradient(zVec),7);

            % Use the smaller of the two
            beta = min(betaDeformation, betaStress);    
            nSamples = nTotalSamples; % total
            zVec = randn(7,nSamples);
            % Only consider samples whose distance is greater than or equal to beta
            
            d = vecnorm(zVec);
            indices  = find(d >= beta);
            
            nReducedSamples = numel(indices);
            disp(['nReducedSamples = ' num2str(nReducedSamples)])
            g1 = zeros(1,numel(indices));
            g2 = zeros(1,numel(indices));
            
            % Evaluate the two g functions outside the circle
            for i = 1:numel(indices)
                g1(i) = obj.gDeformation(zVec(:,indices(i)));
                g2(i) = obj.gStress(zVec(:,indices(i)));
            end
            
            % Composition
            g = g1 + g2 - sqrt(g1.^2 + g2.^2);  
           
            Pf_SE = sum(g<0)/nSamples; % probability of failure
            disp(['Pf_DeformationOrStress = ' num2str(Pf_SE)]);
            disp(['nFunctionEvals = ' num2str(2*length(indices))]);
        end
        function [beta, zStar, Pf_HL,nFunctionEvals,flag]  = gDeformationHL(obj)
            [beta, zStar, Pf_HL,nFunctionEvals,flag] =HansoferLindND(@(zVec)obj.gDeformationWithGradient(zVec),7);
        end
        function [g,gradZ] = gDeformationWithGradient(obj,z)
            h = 1e-6; % step size for finite difference
            g = obj.gDeformation(z);
            N = numel(z);
            gradZ = zeros(1,N);  
            if (nargout == 1)
                return; % no need to compute gradient
            end
            z0 = z; % save the value
            for i = 1:N
                z = z0;
                z(i) = z(i) + h; % increment the i-th value of z
                gh = obj.gDeformation(z);
                gradZ(i) = (gh-g)/h;
            end
        end
        function [g] = gDeformation(obj,z)
            % Reliability function for a specific truss problem
            ANominal = obj.myAreaNominal;
            t = obj.truss7Bar(ANominal); % nominal model
            AStdDev = obj.myAreaStdDev;
            for i = 1:numel(z)
                A =  AStdDev*z(i) + ANominal;
                t = t.assignA(A,i);
            end
            t = t.assemble();
            t = t.solve();
            g = obj.myDeformationSafetyFactor - t.myDeformationMax/obj.myMaxDeformationNominal;
        end
        function [beta, zStar, Pf_HL,nFunctionEvals,flag]  = gStressHL(obj)       
            [beta, zStar, Pf_HL,nFunctionEvals,flag] =HansoferLindND(@(zVec)obj.gStressWithGradient(zVec),7);
        end
        function [g,gradZ] = gStressWithGradient(obj,z)
            h = 1e-6; % step size for finite difference
            g = obj.gStress(z);
            N = numel(z);
            gradZ = zeros(1,N);  
            if (nargout == 1)
                return; % no need to compute gradient
            end
            z0 = z; % save the value
            for i = 1:N
                z = z0;
                z(i) = z(i) + h; % increment the i-th value of z
                gh = obj.gStress(z);
                gradZ(i) = (gh-g)/h;
            end
        end
        function [g] = gStress(obj,z)
            % Reliability function for a specific truss problem
            ANominal = obj.myAreaNominal;
            t = obj.truss7Bar(ANominal); % nominal model
            AStdDev = obj.myAreaStdDev;
            for i = 1:numel(z)
                A =  AStdDev*z(i) + ANominal;
                t = t.assignA(A,i);
            end
            t = t.assemble();
            t = t.solve();
            g = (obj.myStressSafetyFactor- t.myStressMax/obj.myMaxStressNominal);
        end
        function t = truss7Bar(~,Areas)
            xy = [0.5 1.5 1.0 0 2.0; 1 1 0 0 0];% nodes
            connectivity = [1 2; 1 3; 1 4; 2 3; 2 5; 3 4; 3 5]'; % connectivity
            t = truss2d(xy,connectivity);
            t = t.assignE(2e11);
            t = t.assignA(Areas);
            t = t.fixXofNodes([4 5]);
            t = t.fixYofNodes([4 5]);
            t = t.applyForce(1,[1;-2]);
            t = t.applyForce(2,[2;0]);
        end
    end
end
